/*
    WaterSurface class

    An OpenSceneGraph viewer for pyVolution .sww files.
    copyright (C) 2004 Geoscience Australia
*/


#include <watersurface.h>
#include <osg/BlendFunc>
#include <osg/PolygonMode>
#include <osg/ShapeDrawable>
#include <osg/Texture2D>
#include <osg/TexEnv>
#include <osg/TexGen>

#include <osgDB/Registry>
#include <osgDB/ReadFile>


#define DEF_WATER_COLOUR        0, 0, 0.5, 1            // R, G, B, Alpha (blue)
#define DEF_WATER_TRANSPARENCY  0.1                   // 0=opaque, 1=transparent


// constructor
WaterSurface::WaterSurface(SWWReader* sww)
{
    // persistent
    _sww = sww;
    _node = new osg::Geode;
    _geom = new osg::Geometry;
    _stateset = new osg::StateSet;
    _material = new osg::Material;

    // construct local scenegraph hierarchy
    _node->setName("watersurface");
    _node->addDrawable(_geom);
    _node->setStateSet(_stateset);
    _stateset->setAttribute(_material);

    // why???
    _geom->setUseDisplayList(false);

    // initial geometry
    setTimeStep(1);

    _material->setDiffuse(osg::Material::FRONT, osg::Vec4(DEF_WATER_COLOUR));
    _material->setSpecular(osg::Material::FRONT, osg::Vec4(1,1,1,1));
    _material->setTransparency(osg::Material::FRONT_AND_BACK, DEF_WATER_TRANSPARENCY);
    _material->setShininess(osg::Material::FRONT, 128);

    // environment map
    osg::Texture2D* texture = new osg::Texture2D;
    texture->setDataVariance(osg::Object::DYNAMIC);
    texture->setBorderColor(osg::Vec4(1.0f,1.0f,1.0f,0.5f));
    texture->setWrap(osg::Texture::WRAP_S, osg::Texture::REPEAT);
    texture->setImage(osgDB::readImageFile("envmap.jpg"));
    _stateset->setTextureAttributeAndModes( 1, texture, osg::StateAttribute::ON );
    _stateset->setMode( GL_LIGHTING, osg::StateAttribute::OFF );

    // surface transparency
    osg::BlendFunc* osgBlendFunc = new osg::BlendFunc();
    osgBlendFunc->setFunction(osg::BlendFunc::SRC_ALPHA, osg::BlendFunc::ONE_MINUS_SRC_ALPHA);
    _stateset->setAttribute(osgBlendFunc);
    _stateset->setMode(GL_BLEND, osg::StateAttribute::ON);
    _stateset->setRenderingHint(osg::StateSet::TRANSPARENT_BIN);
   
    // automatically generate texture coords
    osg::TexGen* texgen = new osg::TexGen;
    texgen->setMode( osg::TexGen::SPHERE_MAP );

    osg::TexEnv* texenv = new osg::TexEnv;
    texenv->setMode( osg::TexEnv::DECAL );
    texenv->setColor( osg::Vec4(0.6f,0.6f,0.6f,0.2f) );
    _stateset->setTextureAttributeAndModes( 1, texgen, osg::StateAttribute::ON );
    _stateset->setTextureAttribute( 1, texenv );
}



WaterSurface::~WaterSurface()
{
}



void WaterSurface::setTimeStep(unsigned int ts)
{
    // delete if exists
    if( _geom->getNumPrimitiveSets() )
      _geom->removePrimitiveSet(0);  // reference counting actually does the delete

    // local reference to raw height field data
    osg::ref_ptr<osg::Vec3Array> vertices = _sww->getStageVertexArray(ts);
    osg::ref_ptr<osg::UIntArray> indices = _sww->getStageIndexArray();
    osg::ref_ptr<osg::Vec3Array> normals = _sww->getStagePrimitiveNormalArray();
    osg::ref_ptr<osg::Vec3Array> vertexnormals = _sww->getStageVertexNormalArray();

    _geom->setVertexArray( vertices.get() );
    _geom->setVertexIndices( indices.get() );
    _geom->addPrimitiveSet( new osg::DrawArrays(
           osg::PrimitiveSet::TRIANGLES, 0, _sww->getNumberOfStageIndices() ) );

    //_geom->setNormalArray( normals.get() );
    //_geom->setNormalBinding( osg::Geometry::BIND_PER_PRIMITIVE );

    _geom->setNormalArray( vertexnormals.get() );
    _geom->setNormalIndices( indices.get() );
    _geom->setNormalBinding(osg::Geometry::BIND_PER_VERTEX);

}



void WaterSurface::toggleWireframe()
{
    osg::PolygonMode* polyModeObj = 
        dynamic_cast<osg::PolygonMode*>(_stateset->getAttribute(osg::StateAttribute::POLYGONMODE));

    if (!polyModeObj) 
    {
	    polyModeObj = new osg::PolygonMode;
	    _stateset->setAttribute(polyModeObj);
    }

    switch( polyModeObj->getMode(osg::PolygonMode::FRONT_AND_BACK) )
    {
    
      case osg::PolygonMode::FILL : polyModeObj->setMode(osg::PolygonMode::FRONT_AND_BACK,
							 osg::PolygonMode::LINE); break;
      case osg::PolygonMode::LINE : polyModeObj->setMode(osg::PolygonMode::FRONT_AND_BACK,
							 osg::PolygonMode::FILL); break;
      case osg::PolygonMode::POINT : break;
    }
}
